/*
 * FreeCamera.cs
 * 
 * This file is part of the QuickStart Game Engine.  See http://www.codeplex.com/QuickStartEngine
 * for license details.
 */

using System;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;

namespace QuickStart.Entities
{
    /// <summary>
    /// <see cref="FreeCamera"/> is a free-moving camera. Free camera's can be "flown" around as you please using
    /// input, and have little restriction on their movement.
    /// </summary>
    public class FreeCamera : Camera
    {
        private InputPollingHandler inputs;

        /// <summary>
        /// Thumbstick Rotation sensitivity
        /// </summary>
        private const float thumbstickRotModifier = 0.002f;

        /// <summary>
        /// Thumbstick Movement sensitivity
        /// </summary>
        private const float thumbstickMoveModifier = 0.3f;

        /// <summary>
        /// Does the camera handle input inverted?
        /// </summary>
        private bool cameraRotInverted = true;

        private int speed = 1;

        /// <summary>
        /// Current zoom level, 1 would be (1x), 2 would be (2x), etc...
        /// </summary>
        private int zoomLevel = 1;

        /// <summary>
        /// Holds the default field of view. Storing the default FOV allows you to make changes to the camera's
        /// FOV and return back to your original values.
        /// </summary>
        private float defaultFOV = QSConstants.DefaultFOV;

        private Vector3 flatFront = Vector3.Zero;
        private Vector3 tiltedFront = Vector3.Zero;

        /// <summary>
        /// Rotation matrix which contains the pitch rotation data for this camera.
        /// </summary>
        private Matrix pitchMatrix = Matrix.Identity;      // Public for performance reasons

        /// <summary>
        /// Rotation matrix which contains the turn rotation data for this camera.
        /// </summary>
        private Matrix turnMatrix = Matrix.Identity;      // Public for performance reasons

        /// <summary>
        /// Sets camera's pitch (rotation along camera's right vector)
        /// </summary>
        public float Pitch
        {
            get { return this.pitch; }
            set
            {
                this.pitch = value;
                // Set hasChanged flag to true (if we use a bool)
            }
        }
        private float pitch = 0.0f;

        /// <summary>
        /// Sets camera's turn (rotation along camera's up vector)
        /// </summary>
        public float Turn
        {
            get { return this.turn; }
            set
            {
                this.turn = value;
                // Set hasChanged flag to true (if we use a bool)
            }
        }
        private float turn = 0.0f;                                        // Along Z axis (Up axis)

        /// <summary>
        /// Creates a free moving camera.
        /// </summary>
        /// <param name="FOV">Field-of-view, in degrees</param>
        /// <param name="screenWidth">Screen's width (resolution in pixels).</param>
        /// <param name="screenHeight">Screen's height (resolution in pixels).</param>
        /// <param name="nearPlane">Near plane distance</param>
        /// <param name="farPlane">Far plane distance</param>
        /// <param name="game">The <see cref="QSGame"/> the camera belongs to</param>
        public FreeCamera(QSGame game, float FOV, float screenWidth, float screenHeight, float nearPlane, float farPlane)
            : base(game, FOV, screenWidth, screenHeight, nearPlane, farPlane)
        {
            this.cameraType = CameraType.Free;   // This must not be removed
            this.defaultFOV = FOV;
            this.Game.GameMessage += this.Game_GameMessage;
            this.inputs = new InputPollingHandler(game);

            InitButtonListeners();
        }

        /// <summary>
        /// Initialize button listeners. This determines which buttons the input poller will listen for
        /// </summary>
        public void InitButtonListeners()
        {
            // Listen for these keyboard keys
            inputs.AddInputListener(Keys.Up);
            inputs.AddInputListener(Keys.Down);
            inputs.AddInputListener(Keys.Left);
            inputs.AddInputListener(Keys.Right);
            inputs.AddInputListener(Keys.W);
            inputs.AddInputListener(Keys.S);
            inputs.AddInputListener(Keys.A);
            inputs.AddInputListener(Keys.D);

            // Listen for this gamepad button, from this player (example only, Button A isn't set to perform any action in this cam
            inputs.AddInputListener(Buttons.A, PlayerIndex.One);

            // Listen for these mouse buttons, also only for example, this camera uses event-based input for mouse buttons
            inputs.AddInputListener(MouseButton.Left);
            inputs.AddInputListener(MouseButton.Right);
        }

        /// <summary>
        /// Handles mouse move messages
        /// </summary>
        /// <param name="distance">Distance the mouse has moved expressed through a <see cref="Vector2"/></param>
        private void HandleMouseMove(Vector2 distance)
        {
            this.PitchCamera(distance.Y);
            this.YawCamera(-distance.X);
        }

        /// <summary>
        /// Handles mouse button up messages
        /// </summary>
        /// <param name="mouseButton">The <see cref="MouseButton"/> that was released</param>
        private void HandleButtonUp(MouseButton mouseButton)
        {
            switch (mouseButton)
            {
                case MouseButton.Right:
                    this.speed /= 4;
                    break;
            }
        }

        /// <summary>
        /// Handles mouse button down messages
        /// </summary>
        /// <param name="mouseButton">The <see cref="MouseButton"/> that was pressed</param>
        private void HandleButtonDown(MouseButton mouseButton)
        {
            switch (mouseButton)
            {
                case MouseButton.Right:
                    this.speed *= 4;
                    break;
            }
        }

        /// <summary>
        /// Handles the mouse scroll event
        /// </summary>
        /// <param name="distance">Distance the mouse has scrolled</param>
        private void HandleMouseScroll(int distance)
        {
            if (distance > 0)
            {
                this.ZoomIn();
            }
            else
            {
                this.ZoomOut();
            }
        }

        /// <summary>
        /// Message listener for free camera
        /// </summary>
        /// <param name="message">Incoming message</param>
        private void Game_GameMessage(IMessage message)
        {
            switch (message.Type)
            {
                case MessageType.MouseDown:
                    Message<MouseButton> mouseDownMessage = message as Message<MouseButton>;
                    if (mouseDownMessage == null)
                    {
                        throw new ArgumentException("Passed message was not a Message<MouseButton>");
                    }
                    HandleButtonDown(mouseDownMessage.Data);
                    break;

                case MessageType.MouseUp:
                    Message<MouseButton> mouseUpMessage = message as Message<MouseButton>;
                    if (mouseUpMessage == null)
                    {
                        throw new ArgumentException("Passed message was not a Message<MouseButton>");
                    }
                    HandleButtonUp(mouseUpMessage.Data);
                    break;

                case MessageType.MouseScroll:
                    Message<int> scrollWheelMessage = message as Message<int>;
                    if (scrollWheelMessage == null)
                    {
                        throw new ArgumentException("Passed message was not a Message<int>");
                    }
                    HandleMouseScroll(scrollWheelMessage.Data);
                    break;

                case MessageType.MouseMove:
                    Message<Vector2> moveMessage = message as Message<Vector2>;
                    if (moveMessage == null)
                    {
                        throw new ArgumentException("Passed message was not a Message<Vector2>");
                    }
                    HandleMouseMove(moveMessage.Data);
                    break;
            }
        }

        /// <summary>
        /// Process camera movements based on input
        /// </summary>
        /// <param name="gameTime"></param>
        private void ProcessCameraInput(GameTime gameTime)
        {
            float timeDelta = gameTime.ElapsedGameTime.Milliseconds;
            if (timeDelta > 100)
            {
                timeDelta = 100;
            }

            if (inputs.IsDown(Keys.Up))
            {
                PitchCamera(0.001f * timeDelta);
            }

            if (inputs.IsDown(Keys.Down))
            {
                PitchCamera(-0.001f * timeDelta);
            }

            if (inputs.IsDown(Keys.Left))
            {
                YawCamera(0.001f * timeDelta);
            }

            if (inputs.IsDown(Keys.Right))
            {
                YawCamera(-0.001f * timeDelta);
            }

            if (inputs.IsDown(Keys.W))
            {
                MoveFrontBack(0.15f * timeDelta * this.speed);
            }

            if (inputs.IsDown(Keys.S))
            {
                MoveFrontBack(-0.15f * timeDelta * this.speed);
            }

            //if (inputs.IsHeld(Buttons.A, PlayerIndex.One))
            //{
            //    MoveFrontBack(-0.15f * timeDelta * this.speed);
            //}

            if (inputs.IsDown(Keys.A))
            {
                Strafe(-0.15f * timeDelta * this.speed);
            }

            if (inputs.IsDown(Keys.D))
            {
                Strafe(0.15f * timeDelta * this.speed);
            }

            // ===== Thumbstick input =====
            // Left thumbstick moves the camera's position
            Vector2 LeftStick;
            Vector2 RightStick;
            inputs.Thumbstick(PlayerIndex.One, GamePadInputSide.Left, out LeftStick);
            inputs.Thumbstick(PlayerIndex.One, GamePadInputSide.Right, out RightStick);

            if (Math.Abs(LeftStick.Y) > 0.0f)
            {
                this.MoveFrontBack(LeftStick.Y * timeDelta * thumbstickMoveModifier);
            }

            if (Math.Abs(LeftStick.X) > 0.0f)
            {
                this.Strafe(LeftStick.X * timeDelta * thumbstickMoveModifier);
            }

            if (Math.Abs(RightStick.Y) > 0.0f)
            {
                int invertMod = (cameraRotInverted) ? -1 : 1;
                this.PitchCamera(RightStick.Y * timeDelta * thumbstickRotModifier * -invertMod);
            }

            if (Math.Abs(RightStick.X) > 0.0f)
            {
                this.YawCamera(RightStick.X * timeDelta * thumbstickRotModifier);
            }
        }

        /// <summary>
        /// Update the camera. Update the view matrix and frustum if needed. Take into account any movement of the camera.
        /// </summary>
        /// <param name="gameTime">Contains timer information</param>
        public override void Update(GameTime gameTime)
        {
            ProcessCameraInput(gameTime);

            this.Right = Vector3.Cross(this.Up, this.Forward);
            this.flatFront = Vector3.Cross(this.Right, this.Up);

            this.pitchMatrix = Matrix.CreateFromAxisAngle(this.Right, this.Pitch);
            this.turnMatrix = Matrix.CreateFromAxisAngle(this.Up, this.Turn);

            Matrix pitchTurnMat = this.pitchMatrix * this.turnMatrix;
            this.tiltedFront = Vector3.TransformNormal(this.Forward, pitchTurnMat);

            float frontDot = 0.0f;
            Vector3.Dot(ref this.tiltedFront, ref this.flatFront, out frontDot);

            // Check angle so we can't flip over
            if (frontDot > 0.001f)
            {
                this.Forward = Vector3.Normalize(this.tiltedFront);
            }

            // Compute view matrix
            this.viewMatrix = Matrix.CreateLookAt(this.Position, this.Position + this.Forward, this.Up);

            // @todo: Temporary, this should be updated with a flag of some sort. - N.Foster
            this.UpdateFrustum();

            // Reset pitch and turn values so the next input loop can start fresh
            this.Pitch = 0.0f;
            this.Turn = 0.0f;

            // Set hasChanged back to false...
        }

        /// <summary>
        /// Strafes camera left or right (along its right vector)
        /// </summary>
        /// <param name="StrafeValue">Left is negative value, right is positive. 
        /// Value ranges (-1 to 1)</param>
        public void Strafe(float strafeValue)
        {
            this.Position -= this.Right * strafeValue;
        }

        /// <summary>
        /// Moves camera up or down
        /// </summary>
        /// <param name="UpDownValue">Up is negative value, down is positive
        /// Value ranges (-1 to 1)</param>
        public void MoveUpDown(float upDownValue)
        {
            this.Position += this.Up * upDownValue;
        }

        /// <summary>
        /// Moves camera forwards or backwards
        /// </summary>
        /// <param name="FrontBackValue">Forwards is a positive value, backwards is negative.
        /// Values range from -1 to 1</param>
        public void MoveFrontBack(float frontBackValue)
        {
            this.Position += this.Forward * frontBackValue;
        }

        /// <summary>
        /// Yaws camera (rotates along Z axis). Represents a yaw rotation.
        /// </summary>
        /// <param name="RotationValue">Amount to yaw</param>
        public void YawCamera(float rotationValue)
        {
            this.turn += rotationValue;
        }

        /// <summary>
        /// Pitches camera vertically (along its right axis). Represents a pitch rotation.
        /// </summary>
        /// <param name="RotationValue">Amount to pitch</param>
        public void PitchCamera(float rotationValue)
        {
            this.Pitch += rotationValue;
        }

        /// <summary>
        /// Zooms in on current view by 2x the current value (by restricting the <see cref="FOV"/>).
        /// Returns back to default zoom when zoomed past the <see cref="FreeCamera.MaxZoomLevel"/>
        /// </summary>
        public void ZoomIn()
        {
            if (this.zoomLevel < QSConstants.MaxZoomLevel)
            {
                this.zoomLevel *= 2;
            }

            this.FOV = this.defaultFOV / this.zoomLevel;

            this.UpdateProjMatrix();
        }

        /// <summary>
        /// Zooms out on current view (by widening the <see cref="FOV"/>)
        /// </summary>
        public void ZoomOut()
        {
            if (this.zoomLevel > 1)
            {
                this.zoomLevel /= 2;

                this.FOV = this.defaultFOV / this.zoomLevel;

                this.UpdateProjMatrix();
            }
        }
    }
}
