﻿#region Using Statements
using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using JigLibX.Math;
#endregion

namespace JigLibX.Physics
{
    /// <summary>
    /// Constraints a point on one body to be fixed to a point on another body
    /// </summary>
    public class ConstraintPoint : Constraint
    {
        private const float mMaxVelMag = 20.0f;
        private const float minVelForProcessing = 0.01f;

        private Vector3 body0Pos;
        private Body body0;
        private Vector3 body1Pos;
        private Body body1;
        private float allowedDistance;
        private float timescale;

        // some values that we calculate once in pre_apply
        private Vector3 worldPos; ///< average of the two joint positions
        private Vector3 R0; ///< position relative to body 0 (in world space)
        private Vector3 R1;
        private Vector3 vrExtra; ///< extra vel for restoring deviation

        public ConstraintPoint()
        {
        }

        public ConstraintPoint(Body body0, Vector3 body0Pos, Body body1, Vector3 body1Pos, float allowedDistance, float timescale)
        {
            Initialise(body0, body0Pos, body1, body1Pos, allowedDistance, timescale);

        }

        public void Initialise(Body body0, Vector3 body0Pos, Body body1, Vector3 body1Pos, float allowedDistance, float timescale)
        {
            this.body0Pos = body0Pos;
            this.body1Pos = body1Pos;
            this.body0 = body0;
            this.body1 = body1;

            this.allowedDistance = allowedDistance;
            this.timescale = timescale;

            if (timescale < JiggleMath.Epsilon)
                timescale = JiggleMath.Epsilon;

            if (body0 != null) body0.AddConstraint(this);
            if (body1 != null) body1.AddConstraint(this);
        }

        public override void PreApply(float dt)
        {
            this.Satisfied = false;

            R0 = Vector3.Transform(body0Pos, body0.Orientation);
            R1 = Vector3.Transform(body1Pos, body1.Orientation);

            //MultMatrix33(mR0, mBody0->GetOrientation(), mBody0Pos);
            //MultMatrix33(mR1, mBody1->GetOrientation(), mBody1Pos);

            Vector3 worldPos0 = body0.Position + R0;
            Vector3 worldPos1 = body1.Position + R1;

            worldPos = 0.5f * (worldPos0 + worldPos1);

            // add a "correction" based on the deviation of point 0
            Vector3 deviation = worldPos0 - worldPos1;

            float deviationAmount = deviation.Length();

            if (deviationAmount > allowedDistance)
            {
                vrExtra = ((deviationAmount - allowedDistance) /
                            (deviationAmount * System.Math.Max(timescale, dt))) * deviation;
            }
            else
            {
                vrExtra = Vector3.Zero;
            }
        }

        public override bool Apply(float dt)
        {
            this.Satisfied = true;

            bool body0FrozenPre = !body0.IsActive;
            bool body1FrozenPre = !body1.IsActive;

            //  if (body0FrozenPre && body1FrozenPre)
            //    return false;


            Vector3 currentVel0 = (body0.Velocity + Vector3.Cross(body0.AngVel, R0));
            Vector3 currentVel1 = (body1.Velocity + Vector3.Cross(body1.AngVel, R1));

            // add a "correction" based on the deviation of point 0
            Vector3 Vr = (vrExtra + currentVel0 - currentVel1);

            float normalVel = Vr.Length();

            if (normalVel < minVelForProcessing)
                return false;

            // limit things
            if (normalVel > mMaxVelMag)
            {
                Vr *= mMaxVelMag / normalVel;
                normalVel = mMaxVelMag;
            }

            Vector3 N = Vr / normalVel;

            float numerator = -normalVel;

            float denominator = body0.InvMass + body1.InvMass +
              Vector3.Dot(N, Vector3.Cross(Vector3.Transform(Vector3.Cross(R0, N), body0.WorldInvInertia), R0)) +
              Vector3.Dot(N, Vector3.Cross(Vector3.Transform(Vector3.Cross(R1, N), body1.WorldInvInertia), R1));

            if (denominator < JiggleMath.Epsilon)
                return false;

            Vector3 normalImpulse = (numerator / denominator) * N;

            if (!body0.Immovable)
                body0.ApplyWorldImpulse(normalImpulse, worldPos);

            if (!body1.Immovable)
                body1.ApplyWorldImpulse(-normalImpulse, worldPos);

            body0.SetConstraintsAndCollisionsUnsatisfied();
            body1.SetConstraintsAndCollisionsUnsatisfied();

            this.Satisfied = true;

            return true;
        }

        public override void Destroy()
        {
            body0 = null;
            body1 = null;

            DisableConstraint();
        }
    }
}
