﻿#region Using Statements
using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using JigLibX.Collision;
using JigLibX.Math;
#endregion

namespace JigLibX.Physics
{
    /// <summary>
    /// Constrains a point within a rigid body to remain at a fixed world point    
    /// </summary>
    public class ConstraintWorldPoint : Constraint
    {
        const float minVelForProcessing = 0.001f;

        private Body body;
        private Vector3 pointOnBody;
        private Vector3 worldPosition;

        public ConstraintWorldPoint()
        {
            Initialise(null, Vector3.Zero, Vector3.Zero);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="body"></param>
        /// <param name="pointOnBody">pointOnBody is in body coords</param>
        /// <param name="worldPosition"></param>
        public ConstraintWorldPoint(Body body, Vector3 pointOnBody, Vector3 worldPosition)
        {
            Initialise(body, pointOnBody, worldPosition);
        }

        public void Initialise(Body body, Vector3 pointOnBody, Vector3 worldPosition)
        {
            this.body = body;
            this.pointOnBody = pointOnBody;
            this.worldPosition = worldPosition;
         
            if (body!=null) body.AddConstraint(this);
        }


        public override void PreApply(float dt)
        {
            Satisfied = false;
        }

        public override bool Apply(float dt)
        {
            Satisfied = true;

            // transform PointOnBody to the world space
            Vector3 worldPos = body.Position + Vector3.Transform(pointOnBody, body.Orientation);
            //Vector3 worldPos = mBody.Position + mBody.Orientation * mPointOnBody;

            Vector3 R = worldPos - body.Position;
            Vector3 currentVel = body.Velocity + Vector3.Cross(body.AngVel, R);

            // add an extra term to get us back to the original position
            Vector3 desiredVel;

            float allowedDeviation = 0.01f;
            float timescale = 4.0f * dt;
            Vector3 deviation = worldPos - worldPosition;
            float deviationDistance = deviation.Length();

            if (deviationDistance > allowedDeviation)
            {
                Vector3 deviationDir = deviation / deviationDistance;

                desiredVel = ((allowedDeviation - deviationDistance) / timescale) * deviationDir;
            }
            else
            {
                desiredVel = Vector3.Zero;
            }

            // stop velocities pushing us through geometry
            if (body.CollisionSkin != null)
            {
                List<CollisionInfo> collisions = body.CollisionSkin.Collisions;

                int num = collisions.Count;

                for (int i = 0; i < num; i++)
                {
                    CollisionInfo collInfo = collisions[i];

                    if (collInfo.SkinInfo.Skin1.Owner == null)
                    {
                        Vector3 dir = collInfo.DirToBody0;
                        //Vector3 dir = collInfo.DirToBody0.GetNormalisedSafe();

                        float dot = Vector3.Dot(desiredVel, dir);

                        if (dot < 0.0f)
                            desiredVel -= dot * dir;
                    }
                }
            }

            // need an impulse to take us from the current vel to the desired vel
            Vector3 N = currentVel - desiredVel;
            float normalVel = N.Length();

            if (normalVel < minVelForProcessing)
                return false;

            N /= normalVel;

            float denominator = body.InvMass +
                Vector3.Dot(N, Vector3.Cross(Vector3.Transform(Vector3.Cross(R, N), body.WorldInvInertia), R));


            //float denominator = mBody.InvMass +
            //  Vector3.Dot(N, Vector3.Cross(mBody.WorldInvInertia * (Vector3.Cross(R, N)), R));

            if (denominator < JiggleMath.Epsilon)
                return false;

            float normalImpulse = -normalVel / denominator;

            body.ApplyWorldImpulse(normalImpulse * N, worldPos);

            body.SetConstraintsAndCollisionsUnsatisfied();
            Satisfied = true;

            return true;
        }

        public override void Destroy()
        {
            // this is moved here from ConstraintWorldPoint destructor..
            if (body != null) body.RemoveConstraint(this);

            body = null;
            DisableConstraint();

        }

        public Vector3 WorldPosition
        {   
            set { worldPosition = value;}
            get { return worldPosition; }
        }

        public Body Body
        {
            get { return body; }
        }

    }
}
