//
// QuickStartSampleGame.cs
// 
// This file is part of the QuickStart Game Engine.  See http://www.codeplex.com/QuickStartEngine
// for license details.


using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using QuickStart;
using QuickStart.Graphics;
using QuickStart.Entities;
using QuickStart.Compositor;

using QuickStart.Interfaces;
using Microsoft.Xna.Framework.Input;

using QuickStart.Physics;


namespace QuickStartSampleGame
{
    /// <summary>
    /// Sample QuickStart Game implementation
    /// </summary>
    public class QuickStartSampleGame : QSGame
    {
        private SceneManager sceneManager;

        /// <summary>
        /// Sample compositor screens
        /// </summary>
        private FPSScreen fpsScreen;
        
        private QuickStartLogoScreen logoScreen;

        /// <summary>
        /// Input poller so we can query multi-key combos, like CTRL + Enter
        /// </summary>
        private InputPollingHandler inputs;

        /// <summary>
        /// Creates a new instance of the sample game.
        /// </summary>
        public QuickStartSampleGame() 
        {
            this.Content.RootDirectory = "Content";
            this.IsMouseVisible = QSConstants.MouseDefaultVisible;
        }

        /// <summary>
        /// Initializes the current sample game instance.
        /// </summary>
        protected override void Initialize()
        {
            sceneManager = new SceneManager(this);

            // Set up a window with maximum MSAA
            GraphicsSettings settings = new GraphicsSettings();
            settings.BackBufferWidth    = QSConstants.ScreenWidth;
            settings.BackBufferHeight   = QSConstants.ScreenHeight;
            settings.EnableVSync        = QSConstants.IsVSyncd;
            settings.EnableMSAA         = QSConstants.IsMultiSampling;
            settings.EnableFullScreen   = QSConstants.IsFullScreen;

            this.IsFixedTimeStep        = QSConstants.IsFixedTimeStep;

            // Uncomment this line for PIX debugging
            //settings.EnableMSAA = false;

            Graphics.ApplySettings(settings);

            Graphics.RegisterRenderChunkProvider(sceneManager);

            inputs = new InputPollingHandler(this);
            InitInputListeners();

            SendDebugMessage("Initialize");

            // Set up screen compositor
            fpsScreen = new FPSScreen(10, 10);
            Compositor.InsertScreen(fpsScreen, false);
            logoScreen = new QuickStartLogoScreen(false, false, 10, 150);
            Compositor.InsertScreen(logoScreen, false);

            this.Compositor.InsertScreen(new DebugScreen(10, 100, this), false);

            Compositor.InsertScreen(sceneManager, true);

            base.Initialize();
        }        

        /// <summary>
        /// Loads the sample game's content assets.
        /// </summary>
        protected override void LoadContent()
        {
            LoadScenes();

            SendDebugMessage("Load Content");

            base.LoadContent();
        }

        protected override void UnloadContent()
        {
            base.UnloadContent();
        }

        /// <summary>
        /// Add input listeners for input polling
        /// </summary>
        private void InitInputListeners()
        {
            inputs.AddInputListener(Keys.Enter);
            inputs.AddInputListener(Keys.LeftControl);
            inputs.AddInputListener(Keys.RightControl);
            inputs.AddInputListener(Keys.LeftAlt);
            inputs.AddInputListener(Keys.RightAlt);

            inputs.SetLockable(Keys.LeftAlt, true);
            inputs.SetLockable(Keys.RightAlt, true);
        }

        /// <summary>
        /// Temporary method, simply to demonstrate how to load up a scene. Later this
        /// type of thing should be done from an XML file or similar.
        /// </summary>
        private void LoadScenes()
        {
            Scene newScene = new SampleScene(this);

            StaticEntity newEntity = new StaticEntity(this, "Models/Ship", Vector3.Zero, Matrix.Identity);

            newScene.AddEntity(newEntity);

            // Debug message test
            newScene.AddEntity(new DebugMessageEntity(this));

            LoadTerrain(newScene);
            LoadLight(newScene);

            sceneManager.LoadScene(newScene, true);
        }

        /// <summary>
        /// Load up some test terrain.
        /// TEMPORARY!
        /// </summary>        
        private void LoadTerrain(Scene targetScene)
        {
            Terrain newTerrain = new Terrain(this, LOD.High);

            newTerrain.ElevationStrength = 25; 

            newTerrain.InitTerrainTextures("./Textures/grass", "./Textures/rock", "./Textures/sand");

            if (QSConstants.DetailLevel >= GraphicsLevel.High)
                newTerrain.InitTerrainNormalsTextures("./Textures/grassNormal", "./Textures/rockNormal", "./Textures/sandNormal");

            newTerrain.Initialize("./Images/Heightmaps/gcanyon", "./Images/Terrainmaps/gcanyonTerrain", 4, 25, targetScene.PhysicsScene);

            // Sets up the LOD arrays to allow for new LODs later on.
            newTerrain.SetupLODS();

            // Setup a very low LOD terrain. This will allow the water
            // component to use a very low detail terrain for faster
            // rendering. This terrain is not the default unless specified
            // seperately, so the terrain will continue to render at
            // its default setting, this merely sets up a second LOD terrain
            // for use by other components.
            //newTerrain.AddNew(LOD.Low);

            targetScene.sceneTerrain = newTerrain;
        }

        /// <summary>
        /// Load up a light for testing.
        /// TEMPORARY!
        /// </summary>
        /// <param name="targetScene"></param>
        private void LoadLight(Scene targetScene)
        {
            Light newLight = new Light(this);

            newLight.direction = Vector3.Normalize(new Vector3(.5f, -.2f, -.5f));          // Direction this light points
            newLight.setAmbientColor = new Vector3(0.3f, 0.3f, 0.3f);                    // RGB values of our light
            newLight.specularColor = new Vector3(1.0f, 1.0f, 1.0f);
            newLight.SpecularPower = 8f;
            newLight.diffuseColor = new Vector3(1.0f, 1.0f, 1.0f);
            newLight.MinimumAmbient = 0.25f;

            targetScene.sceneLight = newLight;
        }

        /// <summary>
        /// Updates the sample game for the current cycle.
        /// </summary>
        /// <param name="gameTime">A snapshot of the game's time.</param>
        protected override void Update(GameTime gameTime)
        {
            sceneManager.Update(gameTime);

            ProcessInput();

            base.Update(gameTime);
        }

        /// <summary>
        /// Process polling-based input
        /// </summary>
        private void ProcessInput()
        {
            if (inputs.IsDown(Keys.Enter))
            {
                if (inputs.IsDown(Keys.LeftControl) || inputs.IsDown(Keys.RightControl))
                {
                    if (inputs.IsDown(Keys.LeftAlt) || inputs.IsDown(Keys.RightAlt))
                    {
                        this.Graphics.ToggleFullscreen();
                    }
                }
            }
        }

        /// <summary>
        /// Renders a single frame for the sample game.
        /// </summary>
        /// <param name="gameTime">A snapshot of the game's time.</param>
        protected override void Draw(GameTime gameTime)
        {
            // Render the whole frame with one call!
            Compositor.DrawCompositorChain(gameTime);

            base.Draw(gameTime);
        }

        /// <summary>
        /// Handles game messages
        /// </summary>
        /// <param name="message">The <see cref="IMessage"/> send to the game</param>
        protected override void OnGameMessage(IMessage message)
        {
            switch (message.Type)
            {
                case MessageType.KeyDown:
                    KeyMessage msg = message as KeyMessage;
                    switch (msg.Data.Key)
                    {
                        case Keys.Escape:
                            this.SendMessage(new ExitMessage());
                            break;
                    }
                    break;
            }

            base.OnGameMessage(message);
        }

        /// <summary>
        /// Sends a text message to the debug output
        /// </summary>
        /// <param name="message">Text to send to the debug output</param>
        private void SendDebugMessage(string message)
        {
            Message<string> msg = ObjectPool.Aquire<Message<string>>();
            msg.Data = message;
            msg.Type = SampleMessageType.DebugText;

            this.SendMessage(msg);
        }
    }
}
