//
// SampleScene.cs
// 
// This file is part of the QuickStart Game Engine.  See http://www.codeplex.com/QuickStartEngine
// for license details.
//

using System;
using System.Collections.Generic;

using Microsoft.Xna.Framework;

using QuickStart;
using QuickStart.Physics;
using QuickStart.Graphics;

namespace QuickStartSampleGame
{
    /// <summary>
    /// A test scene demonstrating the physics of QuickStart.
    /// </summary>
    public class SampleScene : Scene
    {
        // Physics Temp
        private StaticModel boxModel;
        private StaticModel sphereModel;
        private Material material;
        private readonly List<IPhysicsActor> actors;

        public SampleScene(QSGame game) : base(game)
        {
            this.actors = new List<IPhysicsActor>();

            // Physics Temp
            BoxGrid(10, 20.0f, new Vector3(900.0f, 400.0f, 900.0f));
            
            BoxShapeDesc shape = new BoxShapeDesc();
            shape.Extents = new Vector3(40.0f, 1.0f, 40.0f);

            ActorDesc desc = new ActorDesc();
            desc.Density = 10.0f;
            desc.Dynamic = false;
            desc.Orientation = Matrix.Identity;
            desc.Position = new Vector3(200.0f, 440.0f, 200.0f);
            desc.Shapes.Add(shape);

            IPhysicsActor actor = this.PhysicsScene.CreateActor(desc);

            actors.Add(actor);
        }

        /// <summary>
        /// Cleans up the SampleScene instance for deletion.
        /// </summary>
        ~SampleScene()
        {
            foreach(IPhysicsActor actor in this.actors)
            {
                actor.Dispose();
            }

            actors.Clear();
        }

        /// <summary>
        /// Creates a grid of spheres
        /// </summary>
        /// <param name="size">Grid size (side length)  size^2 spheres will be created.</param>
        /// <param name="spacing">Space between spheres.</param>
        /// <param name="offset">World position of sphere at (0, 0) in grid.</param>
        public void SphereGrid(int size, float spacing, Vector3 offset)
        {
            for(int i = 0; i < size; ++i)
            {
                for(int j = 0; j < size; ++j)
                {
                    ActorDesc desc = new ActorDesc();
                    desc.Orientation = Matrix.Identity;
                    desc.Density = 10.0f;
                    desc.Dynamic = true;
                    desc.Position = offset + new Vector3(i * spacing, 0.0f, j * spacing);

                    SphereShapeDesc shapeDesc = new SphereShapeDesc();
                    shapeDesc.Radius = 2.0f;

                    desc.Shapes.Add(shapeDesc);

                    IPhysicsActor actor = this.PhysicsScene.CreateActor(desc);
                    this.actors.Add(actor);
                }
            }
        }

        /// <summary>
        /// Creates a falling stream of spheres.
        /// </summary>
        /// <param name="amount">Number of spheres to create.</param>
        /// <param name="origin">World position of bottom sphere.</param>
        /// <param name="radius">Radius of each sphere.</param>
        /// <param name="spacing">Spacing between spheres (excluding radii).</param>
        public void SphereStream(int amount, Vector3 origin, float radius, float spacing)
        {
            for(int i = 0; i < amount; ++i)
            {
                ActorDesc desc = new ActorDesc();
                desc.Orientation = Matrix.Identity;
                desc.Density = 10.0f;
                desc.Dynamic = true;
                desc.Position = origin + new Vector3(0.0f, i * (2.0f * radius + spacing), 0.0f);

                SphereShapeDesc shapeDesc = new SphereShapeDesc();
                shapeDesc.Radius = radius;

                desc.Shapes.Add(shapeDesc);

                IPhysicsActor actor = this.PhysicsScene.CreateActor(desc);
                this.actors.Add(actor);
            }
        }

        /// <summary>
        /// Creates a grid of boxes
        /// </summary>
        /// <param name="size">Grid size (side length)  size^2 boxes will be created.</param>
        /// <param name="spacing">Space between boxes.</param>
        /// <param name="offset">World position of box at (0, 0) in grid.</param>
        public void BoxGrid(int size, float spacing, Vector3 offset)
        {
            for(int i = 0; i < size; ++i)
            {
                for(int j = 0; j < size; ++j)
                {
                    ActorDesc desc = new ActorDesc();
                    desc.Orientation = Matrix.Identity;
                    desc.Density = 10.0f;
                    desc.Dynamic = true;
                    desc.Position = offset + new Vector3(i * spacing, 0.0f, j * spacing);

                    BoxShapeDesc shapeDesc = new BoxShapeDesc();
                    shapeDesc.Extents = new Vector3(1.0f, 1.0f, 1.0f);

                    desc.Shapes.Add(shapeDesc);

                    IPhysicsActor actor = this.PhysicsScene.CreateActor(desc);
                    this.actors.Add(actor);
                }
            }
        }

        /// <summary>
        /// Creates a falling stream of boxes.
        /// </summary>
        /// <param name="amount">Number of boxes to create.</param>
        /// <param name="origin">World position of bottom box.</param>
        /// <param name="sideExtent">Side extent of each box (cube).  Half of a side length.</param>
        /// <param name="spacing">Spacing between boxes (excluding box extent).</param>
        public void BoxStream(int amount, Vector3 origin, float sideExtent, float spacing)
        {
            for(int i = 0; i < amount; ++i)
            {
                ActorDesc desc = new ActorDesc();
                desc.Orientation = Matrix.Identity;
                desc.Density = 10.0f;
                desc.Dynamic = true;
                desc.Position = origin + new Vector3(0.0f, i * (2.0f * sideExtent + spacing), 0.0f);

                BoxShapeDesc shapeDesc = new BoxShapeDesc();
                shapeDesc.Extents = new Vector3(sideExtent, sideExtent, sideExtent);

                desc.Shapes.Add(shapeDesc);

                IPhysicsActor actor = this.PhysicsScene.CreateActor(desc);
                this.actors.Add(actor);
            }
        }

        /// <summary>
        /// Loads all Content Manager resources.
        /// </summary>
        public override void LoadContent()
        {
            this.boxModel = this.game.Content.Load<StaticModel>("Models/unit_cube");
            this.sphereModel = this.game.Content.Load<StaticModel>("Models/unit_sphere");
            this.material = this.game.Content.Load<Material>("Material/grass");

            base.LoadContent();
        }

        /// <summary>
        /// Releases all Content Manager resources.
        /// </summary>
        public override void UnloadContent()
        {
            base.UnloadContent();
        }

        /// <summary>
        /// Queries the scene for all potentially visible RenderChunks, given a RenderPassDesc descriptor.
        /// </summary>
        /// <param name="desc">A descriptor for the current rendering pass.</param>
        public override void QueryForRenderChunks(ref RenderPassDesc desc)
        {
            // Quick and dirty render routine for test physics actors.
            for(int i = 0; i < this.actors.Count; ++i)
            {
                IPhysicsActor actor = this.actors[i];

                RenderChunk chunk = this.game.Graphics.AllocateRenderChunk();
                chunk.StartIndex = 0;
                chunk.Material = material;
                chunk.VertexStreamOffset = 0;

                if(actor.Shapes[0] is BoxShapeDesc)
                {
                    chunk.Declaration = this.boxModel.Declaration;
                    chunk.Indices = this.boxModel.IndexBuffer;
                    chunk.PrimitiveCount = this.boxModel.PrimitiveCount;
                    chunk.Type = this.boxModel.GeometryType;
                    chunk.VertexCount = this.boxModel.VertexCount;
                    chunk.VertexStreams.Add(this.boxModel.VertexBuffer);

                    chunk.WorldTransform = Matrix.CreateScale((actor.Shapes[0] as BoxShapeDesc).Extents * 2.0f) * actor.Orientation * Matrix.CreateTranslation(actor.Position);
                }
                else if(actor.Shapes[0] is SphereShapeDesc)
                {
                    chunk.Declaration = this.sphereModel.Declaration;
                    chunk.Indices = this.sphereModel.IndexBuffer;
                    chunk.PrimitiveCount = this.sphereModel.PrimitiveCount;
                    chunk.Type = this.sphereModel.GeometryType;
                    chunk.VertexCount = this.sphereModel.VertexCount;
                    chunk.VertexStreams.Add(this.sphereModel.VertexBuffer);

                    chunk.WorldTransform = Matrix.CreateScale((actor.Shapes[0] as SphereShapeDesc).Radius) * actor.Orientation * Matrix.CreateTranslation(actor.Position);
                }
                else
                {
                    throw new Exception("Bad physics shape");
                }
            }

            base.QueryForRenderChunks(ref desc);
        }
    }
}
